const { promisify } = require('util');
const { exec } = require('child_process');

const { chain, trim, find, map, compact, toLower } = require('lodash');
const rp = require('request-promise');
const chalk = require('chalk');
const inquirer = require('inquirer');
const inquirerAutocomplete = require('inquirer-autocomplete-prompt');
const fuzzy = require('fuzzy');

const {
  TERM_MAX_WIDTH,
  STATUS_TO_SORT_INDEX,
  getRepresentationForStatus,
} = require('./constants');

// Register the fuzzyfinder plugin with inquirer
inquirer.registerPrompt('autocomplete', inquirerAutocomplete);

/**
 * Run a system command
 * @param {string} command 
 */
const runCommand = async command => {
  const { stdout } = await promisify(exec)(command);
  return stdout;
};

/**
 * Retrieve all the issues matching the filter of the user
 * @param {object} config 
 */
const getMyIssues = async config => {
  const authBasicPayload = Buffer.from(
    `${config.user}:${config.jiraToken}`
  ).toString('base64');

  const branches = await getAllLocalBranches();

  return rp({
    uri: `${config.host}/rest/api/2/search?jql=${config.issuesFilter}`,
    headers: {
      'Content-Type': 'application/json',
      Authorization: `Basic ${authBasicPayload}`,
    },
    json: true,
  })
    .then(data => {
      const issues = chain(data)
        .get('issues')
        .map(issue => ({
          id: issue.key,
          summary: issue.fields.summary,
          description: issue.fields.description,
          type: issue.fields.issuetype.name,
          status: issue.fields.status.name,
          url: `${config.host}/browse/${issue.key}`,
          branch: find(
            branches,
            branchName => branchName.indexOf(issue.key) !== -1
          ),
        }))
        .value();

      return { issues };
    })
    .catch(error => {
      return {
        issues: [],
        error,
      };
    });
};

const findPullRequestUrl = async ({ issue, repo }, config) => {
  const branchPrefix = `${toLower(issue.type)}/${issue.id}/`;
  return rp({
    uri: `https://api.github.com/search/issues?q=${branchPrefix}%3Atype:pr%3Arepo:${repo}`,
    headers: {
      'User-Agent': 'Botify Jira CLI',
      Authorization: `Bearer ${config.githubToken}`
    },
    json: true,
  })
    .then(data => {
      const pullUrl = chain(data)
        .get('items')
        .find(({ title }) => title.indexOf(`${issue.id}`) !== -1)
        .get('html_url')
        .value();

      return pullUrl;
    })
    .catch(error => {
      return null;
    });
};

/**
 * Retrieve the list of all the local git branches (in the current folder)
 */
const getAllLocalBranches = async () => {
  const rawBranchOutput = await runCommand(
    `git for-each-ref --format='%(refname:short)' refs/heads`
  );

  return chain(rawBranchOutput)
    .split('\n')
    .map(trim)
    .value();
};

const getLocalRepo = async () => {
  const rawOutput = await runCommand(
    'git config --get remote.origin.url'
  );

  return rawOutput.match(/git@github\.com:(.*)\.git/)[1];
};

/**
 * Apply an "ellipsis" on a string after a specific length
 * @param {string} string 
 * @param {number} maxWidth 
 */
const truncateAfter = (string, maxWidth) => {
  if (string.length > maxWidth) {
    return string.slice(0, maxWidth - 1) + '…';
  }

  return string;
}

/**
 * Perform a fuzzy search on a list of issues
 * @param {array} issues
 */
const searchIssues = issues => (_, query) => {
  return new Promise(resolve => {
    const results = fuzzy.filter(query || '', issues, {
      extract: ({ name }) => name,
    });

    const filteredIssues = map(results, 'original');

    resolve(filteredIssues);
  });
};

/**
 * Allow the user to pick an issue from a list using either:
 * - His keyboard's arrow keys
 * - A fuzzy search
 * @param {array} options.issues The issues from JIRA
 * @param {object} options.config The user's configuration object
 */
const selectIssue = async ({ issues, config }) => {
  const issuesStrings = chain(issues)
    .sortBy(({ status }) => STATUS_TO_SORT_INDEX[status])
    .map(({ id, summary, status, branch }) => {
      const representation = getRepresentationForStatus(
        status,
        config.useEmojis
      );
      const branchString = chalk.dim(branch || 'No branch.');
      const label = `${representation}  [${id}] ${summary} ${branchString}`;

      return {
        value: id,
        name: chalk.reset(truncateAfter(label, TERM_MAX_WIDTH - 3)),
      };
    })
    .value();

  const { issueId } = await inquirer.prompt({
    type: 'autocomplete',
    name: 'issueId',
    message: 'Which issue are you interested in?',
    source: searchIssues(issuesStrings),
    pageSize: 100,
  });

  return find(issues, { id: issueId });
};

const selectAction = async issue => {
  const { actionType } = await inquirer.prompt({
    type: 'list',
    name: 'actionType',
    message: 'What do you want to do with this issue?',
    choices: compact([
      issue.branch
        ? {
            value: 'checkoutBranch',
            name: 'Checkout the branch',
          }
        : {
            value: 'createBranch',
            name: 'Create a branch',
          },
      { value: 'goToJira', name: 'See issue on Jira' },
      { value: 'goToGithub', name: 'See issue on Github' },
    ]),
  });

  return actionType;
}

module.exports = {
  runCommand,
  getMyIssues,
  findPullRequestUrl,
  getAllLocalBranches,
  getLocalRepo,
  selectIssue,
  selectAction,
};
